#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2017-2026 Sébastien Helleu <flashcode@flashtux.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""Unparse AST tree to generate script.

All languages supported by WeeChat can be used: Python, Perl, Ruby, etc.
"""

# ruff: noqa: B905,T201,UP006,UP007,UP035

import argparse
import ast
import inspect
import io
import os
import select
import sys
from typing import Any, List, Union

sys.dont_write_bytecode = True


class Unparse:
    """Unparse AST to generate script code."""


class UnparsePython(Unparse):
    """Unparse AST to generate Python script code.

    This class is inspired from _Unparser class in cpython:
    https://github.com/python/cpython/blob/main/Lib/ast.py

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def __init__(self, output: io.TextIOBase = sys.stdout) -> None:
        """Initialize Python parser."""
        self.output = output
        self.indent_string = " " * 4
        self._indent_level = 0
        self._prefix = []  # not used in Python, only in other languages
        self.binop = {
            "Add": "+",
            "Sub": "-",
            "Mult": "*",
            "MatMult": "@",
            "Div": "/",
            "Mod": "%",
            "LShift": "<<",
            "RShift": ">>",
            "BitOr": "|",
            "BitXor": "^",
            "BitAnd": "&",
            "FloorDiv": "//",
            "Pow": "**",
        }
        self.unaryop = {
            "Invert": "~",
            "Not": "not ",
            "UAdd": "+",
            "USub": "-",
        }
        self.cmpop = {
            "Eq": "==",
            "NotEq": "!=",
            "Lt": "<",
            "LtE": "<=",
            "Gt": ">",
            "GtE": ">=",
            "Is": "is",
            "IsNot": "is not",
            "In": "in",
            "NotIn": "not in",
        }

    def fill(self, string: str = "") -> None:
        """Add a new line and an indented string."""
        self.add(f"\n{self.indent_string * self._indent_level}{string}")

    def indent(self) -> None:
        """Indent code."""
        self._indent_level += 1

    def unindent(self) -> None:
        """Unindent code."""
        self._indent_level -= 1

    def prefix(self, prefix: Union[str, None]) -> None:
        """Add or remove a prefix from list."""
        if prefix:
            self._prefix.append(prefix)
        else:
            self._prefix.pop()

    def add(self, *args) -> None:  # noqa: ANN002
        """Add string/node(s) to the output file."""
        for arg in args:
            if callable(arg):
                arg()
            elif isinstance(arg, tuple):
                arg[0](*arg[1:])
            elif isinstance(arg, list):
                for item in arg:
                    self.add(item)
            elif isinstance(arg, ast.AST):
                method = getattr(self, f"_ast_{arg.__class__.__name__.lower()}", None)
                if method is None:
                    raise NotImplementedError(arg)
                method(arg)
            elif isinstance(arg, str):
                self.output.write(arg)

    @staticmethod
    def make_list(values: List[Any], sep: str = ", ") -> list:
        """Add multiple values using a custom method and separator."""
        result = []
        for value in values:
            if result:
                result.append(sep)
            result.append(value)
        return result

    @staticmethod
    def is_bool(node: ast.AST) -> bool:
        """Check if the node is a boolean."""
        return isinstance(node, ast.Name) and node.id in ("False", "True")

    @staticmethod
    def is_number(node: ast.AST) -> bool:
        """Check if the node is a number."""
        return (isinstance(node, ast.Constant) and isinstance(node.value, int)) or (
            isinstance(node, ast.UnaryOp) and isinstance(node.op, (ast.UAdd, ast.USub))
        )

    def _ast_alias(self, node: ast.AST) -> None:
        """Add an AST alias in output."""
        # ignore alias
        pass  # noqa: PIE790

    def _ast_arg(self, node: ast.AST) -> None:
        """Add an AST arg in output."""
        prefix = self._prefix[-1] if self._prefix else ""
        self.add(f"{prefix}{node.arg}")

    def _ast_assign(self, node: ast.Assign) -> None:
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            [[target, " = "] for target in node.targets],
            node.value,
        )

    def _ast_attribute(self, node: ast.Attribute) -> None:
        """Add an AST Attribute in output."""
        self.add(node.value, ".", node.attr)

    def _ast_binop(self, node: ast.BinOp) -> None:
        """Add an AST BinOp in output."""
        self.add(
            node.left,
            f" {self.binop[node.op.__class__.__name__]} ",
            node.right,
        )

    def _ast_call(self, node: ast.Call) -> None:
        """Add an AST Call in output."""
        self.add(
            node.func,
            "(",
            self.make_list(node.args),
            ")",
        )

    def _ast_compare(self, node: ast.Compare) -> None:
        """Add an AST Compare in output."""
        self.add(node.left)
        for operator, comparator in zip(node.ops, node.comparators):
            self.add(
                f" {self.cmpop[operator.__class__.__name__]} ",
                comparator,
            )

    def _ast_constant(self, node: ast.Constant) -> None:
        """Add an AST Constant in output."""
        self.add(repr(node.value))

    def _ast_dict(self, node: ast.Dict) -> None:
        """Add an AST Dict in output."""
        self.add(
            "{",
            self.make_list([[key, ": ", value] for key, value in zip(node.keys, node.values)]),
            "}",
        )

    def _ast_expr(self, node: ast.Expr) -> None:
        """Add an AST Expr in output."""
        if not isinstance(node.value, ast.Constant):  # ignore docstrings
            self.add(
                self.fill,
                node.value,
            )

    def _ast_functiondef(self, node: ast.FunctionDef) -> None:
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            self.fill if self._indent_level == 0 else None,
            f"def {node.name}(",
            self.make_list(node.args.args),
            "):",
            self.indent,
            node.body,
            self.unindent,
        )

    def _ast_if(self, node: ast.If) -> None:
        """Add an AST If in output."""
        self.add(
            self.fill,
            "if ",
            node.test,
            ":",
            self.indent,
            node.body,
            self.unindent,
        )
        if node.orelse:
            self.add(
                self.fill,
                "else:",
                self.indent,
                node.orelse,
                self.unindent,
            )

    def _ast_index(self, node: ast.Index) -> None:
        """Add an AST Subscript in output."""
        # note: deprecated since Python 3.9
        self.add(node.value)

    def _ast_import(self, node: ast.AST) -> None:
        """Add an AST Import in output."""
        # ignore import
        pass  # noqa: PIE790

    def _ast_list(self, node: ast.AST) -> None:
        """Add an AST List in output."""
        self.add(
            "[",
            self.make_list(node.elts),
            "]",
        )

    def _ast_module(self, node: ast.Module) -> None:
        """Add an AST Module in output."""
        self.add(node.body)

    def _ast_name(self, node: ast.Name) -> None:
        """Add an AST Name in output."""
        prefix = self._prefix[-1] if self._prefix else ""
        self.add(f"{prefix}{node.id}")

    def _ast_num(self, node: ast.Num) -> None:
        """Add an AST Num in output."""
        self.add(repr(node.n))

    def _ast_pass(self, node: ast.Pass) -> None:  # noqa: ARG002
        """Add an AST Pass in output."""
        self.fill("pass")

    def _ast_return(self, node: ast.Return) -> None:
        """Add an AST Return in output."""
        self.fill("return")
        if node.value:
            self.add(" ", node.value)

    def _ast_str(self, node: ast.Str) -> None:
        """Add an AST Str in output."""
        self._ast_constant(node)

    def _ast_subscript(self, node: ast.Subscript) -> None:
        """Add an AST Subscript in output."""
        self.add(
            node.value,
            "[",
            node.slice,
            "]",
        )

    def _ast_tuple(self, node: ast.Tuple) -> None:
        """Add an AST Tuple in output."""
        self.add(
            "(",
            self.make_list(node.elts),
            "," if len(node.elts) == 1 else None,
            ")",
        )

    def _ast_unaryop(self, node: ast.UnaryOp) -> None:
        """Add an AST UnaryOp in output."""
        self.add(
            self.unaryop[node.op.__class__.__name__],
            node.operand,
        )


class UnparsePerl(UnparsePython):
    """Unparse AST to generate Perl script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def _ast_assign(self, node: ast.Assign) -> None:
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            (self.prefix, "$"),
            [[target, " = "] for target in node.targets],
            (self.prefix, None),
            node.value,
            ";",
        )

    def _ast_attribute(self, node: ast.Attribute) -> None:
        """Add an AST Attribute in output."""
        saved_prefix = self._prefix
        self._prefix = []
        self.add(node.value, "::", node.attr)
        self._prefix = saved_prefix

    def _ast_binop(self, node: ast.BinOp) -> None:
        """Add an AST BinOp in output."""
        if isinstance(node.op, ast.Add) and (not self.is_number(node.left) or not self.is_number(node.right)):
            str_op = "."
        else:
            str_op = self.binop[node.op.__class__.__name__]
        self.add(
            (self.prefix, "$"),
            node.left,
            f" {str_op} ",
            node.right,
            (self.prefix, None),
        )

    def _ast_call(self, node: ast.Call) -> None:
        """Add an AST Call in output."""
        self.add(
            node.func,
            "(",
            (self.prefix, "$"),
            self.make_list(node.args),
            (self.prefix, None),
            ")",
        )

    def _ast_compare(self, node: ast.Compare) -> None:
        """Add an AST Compare in output."""
        self.add(node.left)
        for operator, comparator in zip(node.ops, node.comparators):
            if (
                isinstance(operator, (ast.Eq, ast.NotEq))
                and not self.is_number(node.left)
                and not self.is_bool(node.left)
                and not self.is_number(comparator)
                and not self.is_bool(comparator)
            ):
                custom_cmpop = {
                    "Eq": "eq",
                    "NotEq": "ne",
                }
            else:
                custom_cmpop = self.cmpop
            self.add(
                f" {custom_cmpop[operator.__class__.__name__]} ",
                comparator,
            )

    def _ast_constant(self, node: ast.Constant) -> None:
        """Add an AST Constant in output."""
        if isinstance(node.value, str):
            str_node = node.value.replace("$", "\\$").replace("@", "\\@")
            self.add(f'"{str_node}"')
        elif node.value is None:
            self.add("undef")
        else:
            self.add(repr(node.value))

    def _ast_dict(self, node: ast.Dict) -> None:
        """Add an AST Dict in output."""
        self.add(
            "{",
            self.make_list([[key, " => ", value] for key, value in zip(node.keys, node.values)]),
            "}",
        )

    def _ast_expr(self, node: ast.Expr) -> None:
        """Add an AST Expr in output."""
        if not isinstance(node.value, ast.Constant):  # ignore docstrings
            self.add(
                self.fill,
                node.value,
                ";",
            )

    def _ast_functiondef(self, node: ast.FunctionDef) -> None:
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            f"sub {node.name}",
            self.fill,
            "{",
            self.indent,
        )
        if node.args.args:
            self.add(
                self.fill,
                "my (",
                (self.prefix, "$"),
                self.make_list(node.args.args),
                (self.prefix, None),
                ") = @_;",
            )
        self.add(
            node.body,
            self.unindent,
            self.fill,
            "}",
        )

    def _ast_if(self, node: ast.If) -> None:
        """Add an AST If in output."""
        self.add(
            self.fill,
            "if (",
            (self.prefix, "$"),
            node.test,
            (self.prefix, None),
            ")",
            self.fill,
            "{",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "}",
        )
        if node.orelse:
            self.add(
                self.fill,
                "else",
                self.fill,
                "{",
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                "}",
            )

    def _ast_list(self, node: ast.List) -> None:
        """Add an AST List in output."""
        self.add(
            "(",
            self.make_list(node.elts),
            ")",
        )

    def _ast_pass(self, node: ast.Pass) -> None:
        """Add an AST Pass in output."""
        pass  # noqa: PIE790

    def _ast_return(self, node: ast.Return) -> None:
        """Add an AST Return in output."""
        self.fill("return")
        if node.value:
            self.add(
                " ",
                (self.prefix, "%" if isinstance(node.value, ast.Dict) else "$"),
                node.value,
                (self.prefix, None),
                ";",
            )

    def _ast_str(self, node: ast.Str) -> None:
        """Add an AST Str in output."""
        self._ast_constant(node)

    def _ast_subscript(self, node: ast.Subscript) -> None:
        """Add an AST Subscript in output."""
        self.add(
            (self.prefix, "$"),
            node.value,
            (self.prefix, None),
            "->{",
            node.slice,
            "}",
        )


class UnparseRuby(UnparsePython):
    """Unparse AST to generate Ruby script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def _ast_attribute(self, node: ast.Attribute) -> None:
        """Add an AST Attribute in output."""
        self.add(
            node.value,
            "::" if node.attr.startswith("WEECHAT_") else ".",
            node.attr,
        )

    def _ast_constant(self, node: ast.Constant) -> None:
        """Add an AST Constant in output."""
        if isinstance(node.value, str):
            str_node = node.value.replace("#{", "\\#{")
            self.add(f'"{str_node}"')
        elif node.value is None:
            self.add("nil")
        else:
            self.add(repr(node.value))

    def _ast_dict(self, node: ast.Dict) -> None:
        """Add an AST Dict in output."""
        self.add(
            "Hash[",
            self.make_list([[key, " => ", value] for key, value in zip(node.keys, node.values)]),
            "]",
        )

    def _ast_functiondef(self, node: ast.FunctionDef) -> None:
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            f"def {node.name}",
        )
        if node.args.args:
            self.add(
                "(",
                self.make_list(node.args.args),
                ")",
            )
        self.add(
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "end",
        )

    def _ast_if(self, node: ast.If) -> None:
        """Add an AST If in output."""
        self.add(
            self.fill,
            "if ",
            node.test,
            self.indent,
            node.body,
            self.unindent,
        )
        if node.orelse:
            self.add(
                self.fill,
                "else",
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                "end",
            )

    def _ast_list(self, node: ast.List) -> None:
        """Add an AST List in output."""
        self.add(
            "Array[",
            self.make_list(node.elts),
            "]",
        )

    def _ast_pass(self, node: ast.Pass) -> None:
        """Add an AST Pass in output."""
        pass  # noqa: PIE790

    def _ast_str(self, node: ast.Str) -> None:
        """Add an AST Str in output."""
        self._ast_constant(node)


class UnparseLua(UnparsePython):
    """Unparse AST to generate Lua script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002,ANN003
        """Initialize Lua parser."""
        super().__init__(*args, **kwargs)
        self.cmpop = {
            "Eq": "==",
            "NotEq": "~=",
            "Lt": "<",
            "LtE": "<=",
            "Gt": ">",
            "GtE": ">=",
        }

    def _ast_binop(self, node: ast.BinOp) -> None:
        """Add an AST BinOp in output."""
        if isinstance(node.op, ast.Add) and (not self.is_number(node.left) or not self.is_number(node.right)):
            str_op = ".."
        else:
            str_op = self.binop[node.op.__class__.__name__]
        self.add(
            node.left,
            f" {str_op} ",
            node.right,
        )

    def _ast_constant(self, node: ast.Constant) -> None:
        """Add an AST Constant in output."""
        if node.value is None:
            self.add("nil")
        else:
            self.add(repr(node.value))

    def _ast_dict(self, node: ast.Dict) -> None:
        """Add an AST Dict in output."""
        self.add(
            "{",
            self.make_list([["[", key, "]", "=", value] for key, value in zip(node.keys, node.values)]),
            "}",
        )

    def _ast_functiondef(self, node: ast.FunctionDef) -> None:
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            f"function {node.name}",
        )
        self.add(
            "(",
            self.make_list(node.args.args),
            ")",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "end",
        )

    def _ast_if(self, node: ast.If) -> None:
        """Add an AST If in output."""
        self.add(
            self.fill,
            "if ",
            node.test,
            " then",
            self.indent,
            node.body,
            self.unindent,
        )
        if node.orelse:
            self.add(
                self.fill,
                "else",
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                "end",
            )

    def _ast_list(self, node: ast.List) -> None:
        """Add an AST List in output."""
        self.add(
            "{",
            self.make_list(node.elts),
            "}",
        )

    def _ast_pass(self, node: ast.Pass) -> None:
        """Add an AST Pass in output."""
        pass  # noqa: PIE790


class UnparseTcl(UnparsePython):
    """Unparse AST to generate Tcl script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002,ANN003
        """Initialize Tcl parser."""
        super().__init__(*args, **kwargs)
        self._call = 0

    def _ast_assign(self, node: ast.Assign) -> None:
        """Add an AST Assign in output."""
        exclude_types = (ast.Dict, ast.List, ast.Constant, ast.Subscript)
        self.add(
            self.fill,
            "set ",
            node.targets[0],
            " ",
            "[" if not isinstance(node.value, exclude_types) else "",
            node.value,
            "]" if not isinstance(node.value, exclude_types) else "",
        )

    def _ast_attribute(self, node: ast.Attribute) -> None:
        """Add an AST Attribute in output."""
        saved_prefix = self._prefix
        self._prefix = []
        if node.attr.startswith("WEECHAT_"):
            self.add("$::")
        self.add(node.value, "::", node.attr)
        self._prefix = saved_prefix

    def _ast_binop(self, node: ast.BinOp) -> None:
        """Add an AST BinOp in output."""
        self.add(
            "[join [list ",
            (self.prefix, "$"),
            node.left,
            " ",
            node.right,
            (self.prefix, None),
            '] ""]',
        )

    def _ast_call(self, node: ast.Call) -> None:
        """Add an AST Call in output."""
        if self._call:
            self.add("[")
        self._call += 1
        self.add(
            node.func,
            " " if node.args else None,
            (self.prefix, "$"),
            self.make_list(node.args, sep=" "),
            (self.prefix, None),
        )
        self._call -= 1
        if self._call:
            self.add("]")

    def _ast_compare(self, node: ast.Compare) -> None:
        """Add an AST Compare in output."""
        self.prefix("$")
        if self._call:
            self.add("[expr {")
        self.add(node.left)
        for operator, comparator in zip(node.ops, node.comparators):
            if (
                isinstance(operator, (ast.Eq, ast.NotEq))
                and not self.is_number(node.left)
                and not self.is_bool(node.left)
                and not self.is_number(comparator)
                and not self.is_bool(comparator)
            ):
                custom_cmpop = {
                    "Eq": "eq",
                    "NotEq": "ne",
                }
            else:
                custom_cmpop = self.cmpop
            self.add(
                f" {custom_cmpop[operator.__class__.__name__]} ",
                comparator,
            )
        if self._call:
            self.add("}]")
        self.prefix(None)

    def _ast_constant(self, node: ast.Constant) -> None:
        """Add an AST Constant in output."""
        if isinstance(node.value, str):
            str_node = node.value.replace("$", "\\$")
            self.add(f'"{str_node}"')
        elif node.value is None:
            self.add("$::weechat::WEECHAT_NULL")
        else:
            self.add(repr(node.value))

    def _ast_dict(self, node: ast.Dict) -> None:
        """Add an AST Dict in output."""
        self.add(
            "[dict create ",
            self.make_list([[key, " ", value] for key, value in zip(node.keys, node.values)], sep=" "),
            "]",
        )

    def _ast_functiondef(self, node: ast.FunctionDef) -> None:
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            f"proc {node.name} {{",
            (self.make_list(node.args.args, sep=" ") if node.args.args else None),
            "} {",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "}",
        )

    def _ast_if(self, node: ast.If) -> None:
        """Add an AST If in output."""
        self.add(
            self.fill,
            "if {",
            (self.prefix, "$"),
            node.test,
            (self.prefix, None),
            "} {",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "}",
        )
        if node.orelse:
            self.add(
                " else {",
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                "}",
            )

    def _ast_list(self, node: ast.List) -> None:
        """Add an AST List in output."""
        self.add(
            "[",
            self.make_list(node.elts, sep=" "),
            "]",
        )

    def _ast_pass(self, node: ast.Pass) -> None:
        """Add an AST Pass in output."""
        pass  # noqa: PIE790

    def _ast_return(self, node: ast.Return) -> None:
        """Add an AST Return in output."""
        self.fill("return")
        if node.value:
            self.add(
                " ",
                (self.prefix, "$"),
                node.value,
                (self.prefix, None),
            )

    def _ast_str(self, node: ast.Str) -> None:
        """Add an AST Str in output."""
        self._ast_constant(node)

    def _ast_subscript(self, node: ast.Subscript) -> None:
        """Add an AST Subscript in output."""
        self.add(
            "[dict get ",
            (self.prefix, "$"),
            node.value,
            (self.prefix, None),
            " ",
            node.slice,
            "]",
        )


class UnparseGuile(UnparsePython):
    """Unparse AST to generate Guile script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002,ANN003
        """Initialize Guile parser."""
        super().__init__(*args, **kwargs)
        self.cmpop = {
            "Eq": "=",
            "NotEq": "<>",
            "Lt": "<",
            "LtE": "<=",
            "Gt": ">",
            "GtE": ">=",
        }
        self._call = 0
        self._let = 0

    def _ast_assign(self, node: ast.Assign) -> None:
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            "(let ((",
            node.targets[0],
            " ",
            node.value,
            "))",
            self.indent,
            self.fill,
            "(begin",
            self.indent,
        )
        self._let += 1

    def _ast_attribute(self, node: ast.Attribute) -> None:
        """Add an AST Attribute in output."""
        self.add(node.value, ":", node.attr)

    def _ast_binop(self, node: ast.BinOp) -> None:
        """Add an AST BinOp in output."""
        if isinstance(node.op, ast.Add) and (
            isinstance(node.left, (ast.Name, ast.Constant)) or isinstance(node.right, (ast.Name, ast.Constant))
        ):
            self.add(
                "(string-append ",
                node.left,
                " ",
                node.right,
                ")",
            )
        else:
            self.add(
                node.left,
                f" {self.binop[node.op.__class__.__name__]} ",
                node.right,
            )

    def _ast_call(self, node: ast.Call) -> None:
        """Add an AST Call in output."""
        self._call += 1
        self.add(
            "(",
            node.func,
            " " if node.args else None,
            self.make_list(node.args, sep=" "),
            ")",
        )
        self._call -= 1

    def _ast_compare(self, node: ast.Compare) -> None:
        """Add an AST Compare in output."""
        for operator, comparator in zip(node.ops, node.comparators):
            if (
                isinstance(operator, (ast.Eq, ast.NotEq))
                and not self.is_number(node.left)
                and not self.is_bool(node.left)
                and not self.is_number(comparator)
                and not self.is_bool(comparator)
            ):
                prefix = "string"
            else:
                prefix = ""
            self.add(
                f"({prefix}{self.cmpop[operator.__class__.__name__]} ",
                node.left,
                " ",
                comparator,
                ")",
            )

    def _ast_constant(self, node: ast.Constant) -> None:
        """Add an AST Constant in output."""
        if isinstance(node.value, str):
            self.add(f'"{node.value}"')
        elif node.value is None:
            self.add("#nil")
        else:
            self.add(repr(node.value))

    def _ast_dict(self, node: ast.Dict) -> None:
        """Add an AST Dict in output."""
        self.add(
            "'(",
            self.make_list(
                [["(", key, " ", value, ")"] for key, value in zip(node.keys, node.values)], sep=" ",
            ),
            ")",
        )

    def _ast_functiondef(self, node: ast.FunctionDef) -> None:
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            f"(define ({node.name}",
            " " if node.args.args else None,
            (self.make_list(node.args.args, sep=" ") if node.args.args else None),
            ")",
            self.indent,
            node.body,
        )
        while self._let > 0:
            self.add(
                self.unindent,
                self.fill,
                ")",
                self.unindent,
                self.fill,
                ")",
            )
            self._let -= 1
        self.add(
            self.unindent,
            self.fill,
            ")",
        )

    def _ast_if(self, node: ast.If) -> None:
        """Add an AST If in output."""
        self.add(
            self.fill,
            "(if " if isinstance(node.test, ast.Name) else "(",
            node.test,
            "" if isinstance(node.test, ast.Name) else ")",
            self.indent,
            self.fill,
            "(begin",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            ")",
            self.unindent,
        )
        if node.orelse:
            self.add(
                self.indent,
                self.fill,
                "(begin",
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                ")",
                self.unindent,
            )
        self.add(self.fill, ")")

    def _ast_list(self, node: ast.List) -> None:
        """Add an AST List in output."""
        self.add(
            "'(",
            self.make_list(node.elts, sep=" "),
            ")",
        )

    def _ast_pass(self, node: ast.Pass) -> None:  # noqa: ARG002
        """Add an AST Pass in output."""
        self.add("#t")

    def _ast_return(self, node: ast.Return) -> None:
        """Add an AST Return in output."""
        if node.value:
            self.add(self.fill, node.value)

    def _ast_str(self, node: ast.Str) -> None:
        """Add an AST Str in output."""
        self._ast_constant(node)

    def _ast_subscript(self, node: ast.Subscript) -> None:
        """Add an AST Subscript in output."""
        self.add(
            "(assoc-ref ",
            node.value,
            " ",
            node.slice,
            ")",
        )


class UnparseJavaScript(UnparsePython):
    """Unparse AST to generate JavaScript script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def _ast_dict(self, node: ast.Dict) -> None:
        """Add an AST Dict in output."""
        self.add(
            "{",
            self.make_list([[key, ": ", value] for key, value in zip(node.keys, node.values)]),
            "}",
        )

    def _ast_constant(self, node: ast.Constant) -> None:
        """Add an AST Constant in output."""
        if node.value is None:
            self.add("null")
        else:
            self.add(repr(node.value))

    def _ast_functiondef(self, node: ast.FunctionDef) -> None:
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            f"function {node.name}(",
            self.make_list(node.args.args),
            ") {",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "}",
        )

    def _ast_if(self, node: ast.If) -> None:
        """Add an AST If in output."""
        self.add(
            self.fill,
            "if (",
            node.test,
            ") {",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "}",
        )
        if node.orelse:
            self.add(
                " else {",
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                "}",
            )

    def _ast_pass(self, node: ast.Pass) -> None:
        """Add an AST Pass in output."""
        pass  # noqa: PIE790


class UnparsePhp(UnparsePython):
    """Unparse AST to generate PHP script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def _ast_assign(self, node: ast.Assign) -> None:
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            (self.prefix, "$"),
            [[target, " = "] for target in node.targets],
            (self.prefix, None),
            node.value,
            ";",
        )

    def _ast_attribute(self, node: ast.Attribute) -> None:
        """Add an AST Attribute in output."""
        saved_prefix = self._prefix
        self._prefix = []
        if not node.attr.startswith("WEECHAT_"):
            self.add(node.value, "_")
        self.add(node.attr)
        self._prefix = saved_prefix

    def _ast_binop(self, node: ast.BinOp) -> None:
        """Add an AST BinOp in output."""
        if isinstance(node.op, ast.Add) and (
            isinstance(node.left, (ast.Name, ast.Constant)) or isinstance(node.right, (ast.Name, ast.Constant))
        ):
            str_op = "."
        else:
            str_op = self.binop[node.op.__class__.__name__]
        self.add(
            (self.prefix, "$"),
            node.left,
            f" {str_op} ",
            node.right,
            (self.prefix, None),
        )

    def _ast_call(self, node: ast.Call) -> None:
        """Add an AST Call in output."""
        self.add(
            node.func,
            "(",
            (self.prefix, "$"),
            self.make_list(node.args),
            (self.prefix, None),
            ")",
        )

    def _ast_constant(self, node: ast.Constant) -> None:
        """Add an AST Constant in output."""
        if isinstance(node.value, str):
            str_node = node.value.replace("$", "\\$")
            self.add(f'"{str_node}"')
        elif node.value is None:
            self.add("NULL")
        else:
            self.add(repr(node.value))

    def _ast_dict(self, node: ast.Dict) -> None:
        """Add an AST Dict in output."""
        self.add(
            "array(",
            self.make_list([[key, " => ", value] for key, value in zip(node.keys, node.values)]),
            ")",
        )

    def _ast_expr(self, node: ast.Expr) -> None:
        """Add an AST Expr in output."""
        if not isinstance(node.value, ast.Constant):  # ignore docstrings
            self.add(
                self.fill,
                node.value,
                ";",
            )

    def _ast_functiondef(self, node: ast.FunctionDef) -> None:
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            f"function {node.name}(",
            (self.prefix, "$"),
            self.make_list(node.args.args),
            (self.prefix, None),
            ")",
            self.fill,
            "{",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "}",
        )

    def _ast_list(self, node: ast.List) -> None:
        """Add an AST List in output."""
        self.add(
            "array(",
            self.make_list(node.elts),
            ")",
        )

    def _ast_if(self, node: ast.If) -> None:
        """Add an AST If in output."""
        self.add(
            self.fill,
            "if (",
            (self.prefix, "$"),
            node.test,
            (self.prefix, None),
            ")",
            self.fill,
            "{",
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            "}",
        )
        if node.orelse:
            self.add(
                self.fill,
                "else",
                self.fill,
                "{",
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                "}",
            )

    def _ast_pass(self, node: ast.Pass) -> None:
        """Add an AST Pass in output."""
        pass  # noqa: PIE790

    def _ast_return(self, node: ast.Return) -> None:
        """Add an AST Return in output."""
        self.fill("return")
        if node.value:
            self.add(
                " ",
                (self.prefix, "$"),
                node.value,
                (self.prefix, None),
                ";",
            )

    def _ast_str(self, node: ast.Str) -> None:
        """Add an AST Str in output."""
        self._ast_constant(node)

    def _ast_subscript(self, node: ast.Subscript) -> None:
        """Add an AST Subscript in output."""
        self.add(
            "$",
            node.value,
            "[",
            node.slice,
            "]",
        )


def get_languages() -> List[str]:
    """Return a list of supported languages: ['python', 'perl', ...]."""
    members = [
        member
        for member in inspect.getmembers(sys.modules[__name__], predicate=inspect.isclass)
        if inspect.isclass(member[1]) and member[0].startswith("Unparse") and member[0] != "Unparse"
    ]
    return [name[7:].lower() for name, _ in sorted(members, key=lambda member: member[1].__lineno__)]


LANGUAGES = get_languages()


def get_parser() -> argparse.ArgumentParser:
    """Get parser arguments."""
    all_languages = [*LANGUAGES, "all"]
    default_language = LANGUAGES[0]
    parser = argparse.ArgumentParser(
        description=(
            "Unparse Python code from stdin and generate code in "
            "another language (to stdout).\n\n"
            "The code is read from stdin and generated code is "
            "written on stdout."
        ),
    )
    parser.add_argument(
        "-l",
        "--language",
        default=default_language,
        choices=all_languages,
        help="output language (default: {default_language})",
    )
    return parser


def get_stdin() -> str:
    """Return data from standard input.

    If there is nothing in stdin, wait for data until ctrl-d (EOF)
    is received.
    """
    data = ""
    inr = select.select([sys.stdin], [], [], 0)[0]
    if not inr:
        print("Enter the code to convert (Enter + ctrl-d to end)")
    while True:
        inr = select.select([sys.stdin], [], [], 0.1)[0]
        if not inr:
            continue
        new_data = os.read(sys.stdin.fileno(), 4096)
        if not new_data:  # EOF?
            break
        data += new_data.decode("utf-8")
    return data


def convert_to_language(code: str, language: str, prefix: str = "") -> None:
    """Convert Python code to a language."""
    class_name = f"Unparse{language.capitalize()}"
    unparse_class = getattr(sys.modules[__name__], class_name)
    if prefix:
        print(prefix)
    output = io.StringIO()
    unparse_class(output=output).add(ast.parse(code))
    print(output.getvalue().lstrip())


def convert(code: str, language: str) -> None:
    """Convert Python code to one or all languages."""
    if language == "all":
        for lang in LANGUAGES:
            convert_to_language(code, lang, "\n{lang}:")
    else:
        convert_to_language(code, language)


def main() -> None:
    """Generate scripts."""
    parser = get_parser()
    args = parser.parse_args()

    code = get_stdin()
    if not code:
        print("ERROR: missing input")
        print()
        parser.print_help()
        sys.exit(1)

    convert(code, args.language)

    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2019 Simmo Saan <simmo.saan@gmail.com>
# SPDX-FileCopyrightText: 2021-2026 Sébastien Helleu <flashcode@flashtux.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""Generate Python stub with API constants and functions.

This script requires Python 3.10+.
"""

# ruff: noqa: T201

import re
from pathlib import Path
from textwrap import indent

DOC_DIR = Path(__file__).resolve().parent.parent / "doc" / "en"
SRC_DIR = Path(__file__).resolve().parent.parent / "src"

STUB_HEADER = """\
#
# SPDX-FileCopyrightText: 2019-2026 Sébastien Helleu <flashcode@flashtux.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# WeeChat Python stub file, auto-generated by generate_python_stub.py.
# DO NOT EDIT BY HAND!
#

# ruff: noqa: A002,D400,D205,D301,D415,E501,PIE790,PYI021,PYI048,UP006,UP007,UP035

from typing import Dict, Union
"""

CONSTANT_RE = r"WEECHAT_SCRIPT_CONST_(?P<type>(INT|STR))\((?P<constant>WEECHAT_[A-Z0-9_]+)\)"

FUNCTION_RE = r"""\[source,python\]
----
# prototype
def (?P<function>\w+)(?P<args>[^)]*)(?P<return>\) -> [^:]+:) \.\.\.(?P<example>.*?)
----"""


def print_stub_constants() -> None:
    """Print constants, extracted from the plugin-script.c."""
    constant_pattern = re.compile(CONSTANT_RE)
    plugin_script_c = Path(SRC_DIR / "plugins" / "plugin-script.c")
    weechat_plugin_h = Path(SRC_DIR / "plugins" / "weechat-plugin.h")
    with (
        plugin_script_c.open(encoding="utf-8") as plugin_script_file,
        weechat_plugin_h.open(encoding="utf-8") as plugin_public_header_file,
    ):
        plugin_script = plugin_script_file.read()
        plugin_public_header = plugin_public_header_file.read()
        for match in constant_pattern.finditer(plugin_script):
            value_re = rf'^#define {match["constant"]} +(?P<value>[\w"-]+)$'
            value_match = re.search(value_re, plugin_public_header, re.MULTILINE)
            value = f" = {value_match['value']}" if value_match else ""
            print(f"{match['constant']}: {match['type'].lower()}{value}")


def print_stub_functions() -> None:
    """Print function prototypes, extracted from the Plugin API reference."""
    function_pattern = re.compile(FUNCTION_RE, re.DOTALL)
    with Path(DOC_DIR / "weechat_plugin_api.en.adoc").open(encoding="utf-8") as api_doc_file:
        api_doc = api_doc_file.read()
        for match in function_pattern.finditer(api_doc):
            url = f"https://weechat.org/doc/weechat/api/#_{match['function']}"
            example = f"\n    ::\n\n{indent(match['example'].lstrip(), ' ' * 8)}\n    " if match["example"] else ""
            str_func = f"""\n
def {match["function"]}{match["args"]}{match["return"]}
    \"""`{match["function"]} in WeeChat plugin API reference <{url}>`_{example}\"""
    ..."""
            print(str_func)


def stub_api() -> None:
    """Write Python stub file."""
    print(STUB_HEADER)
    print_stub_constants()
    print_stub_functions()


if __name__ == "__main__":
    stub_api()



#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2017-2026 Sébastien Helleu <flashcode@flashtux.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""Test script generator for WeeChat.

Build source of scripts in all languages to test the scripting API.

This script can be run in WeeChat or as a standalone script
(during automatic tests, it is loaded as a WeeChat script).

It uses the following scripts:
- unparse.py: convert Python code to other languages (including Python itself)
- testapi.py: the WeeChat scripting API tests
"""

# ruff: noqa: T201,UP006,UP007,UP035

import argparse
import ast
import datetime
import inspect
import io
import os
import sys
import traceback
from pathlib import Path
from typing import Tuple, Type, Union

sys.dont_write_bytecode = True

SCRIPT_DIR = Path(__file__).resolve().parent
sys.path.append(str(SCRIPT_DIR))
from unparse import (  # noqa: E402
    Unparse,
    UnparseGuile,
    UnparseJavaScript,
    UnparseLua,
    UnparsePerl,
    UnparsePhp,
    UnparsePython,
    UnparseRuby,
    UnparseTcl,
)

RUNNING_IN_WEECHAT = True
try:
    import weechat
except ImportError:
    RUNNING_IN_WEECHAT = False


SCRIPT_NAME = "testapigen"
SCRIPT_AUTHOR = "Sébastien Helleu <flashcode@flashtux.org>"
SCRIPT_VERSION = "0.1"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC = "Generate scripting API test scripts"

SCRIPT_COMMAND = "testapigen"


class WeechatScript:
    """A generic WeeChat script.

    This class must NOT be instantiated directly, use subclasses instead:
    PythonScript, PerlScript, ...
    """

    def __init__(
        self,
        unparse_class: Type[Unparse],
        tree: ast.AST,
        source_script: str,
        output_dir: str,
        language: str,
        extension: str,
        comment_char: str = "#",
        weechat_module: str = "weechat",
    ) -> None:
        """Initialize a WeeChat script generator."""
        self.unparse_class = unparse_class
        self.tree = tree
        self.source_script = os.path.realpath(source_script)
        self.output_dir = os.path.realpath(output_dir)
        self.language = language
        self.extension = extension
        self.script_name = f"weechat_testapi.{extension}"
        self.script_path = Path(self.output_dir) / self.script_name
        self.comment_char = comment_char
        self.weechat_module = weechat_module
        self.update_tree()

    def comment(self, string: str) -> str:
        """Get a commented line."""
        return f"{self.comment_char} {string}"

    def update_tree(self) -> None:
        """Make changes in AST tree."""
        functions = {
            "prnt": "print",
            "prnt_date_tags": "print_date_tags",
            "prnt_datetime_tags": "print_datetime_tags",
            "prnt_y": "print_y",
            "prnt_y_date_tags": "print_y_date_tags",
            "prnt_y_datetime_tags": "print_y_datetime_tags",
        }
        tests_count = 0
        for node in ast.walk(self.tree):
            # rename some API functions
            if (
                self.language != "python"
                and isinstance(node, ast.Call)
                and isinstance(node.func, ast.Attribute)
                and node.func.value.id == "weechat"
            ):
                node.func.attr = functions.get(node.func.attr, node.func.attr)
            # count number of tests
            if isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id == "check":
                tests_count += 1

        # replace script variables in string values
        variables = {
            "{SCRIPT_SOURCE}": self.source_script,
            "{SCRIPT_NAME}": self.script_name,
            "{SCRIPT_PATH}": str(self.script_path),
            "{SCRIPT_AUTHOR}": "Sebastien Helleu",
            "{SCRIPT_VERSION}": "1.0",
            "{SCRIPT_LICENSE}": "GPL3",
            "{SCRIPT_DESCRIPTION}": f"{self.language.capitalize()} scripting API test",
            "{SCRIPT_LANGUAGE}": self.language,
            "{SCRIPT_TESTS}": str(tests_count),
        }
        # replace variables
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Constant) and node.value in variables:
                node.value = variables[node.value]

    def write_header(self, output: io.TextIOBase) -> None:
        """Generate script header (just comments by default)."""
        comments = (
            "",
            f"{self.script_name} -- WeeChat {self.language.capitalize()} scripting API testing",
            "",
            "WeeChat script automatically generated by testapigen.py.",
            "DO NOT EDIT BY HAND!",
            "",
            f"Date: {datetime.datetime.now(tz=datetime.timezone.utc)}",
            "",
        )
        output.writelines(f"{self.comment(line)}\n" for line in comments)

    def write(self) -> None:
        """Write script on disk."""
        print(f"Writing script {self.script_path}... ", end="")
        with self.script_path.open("w") as output:
            self.write_header(output)
            self.unparse_class(output).add(self.tree)
            output.write("\n")
            self.write_footer(output)
        print("OK")

    def write_footer(self, output: io.TextIOBase) -> None:
        """Write footer (nothing by default)."""


class WeechatPythonScript(WeechatScript):
    """A WeeChat script written in Python."""

    def __init__(self, tree: ast.AST, source_script: str, output_dir: str) -> None:
        """Initialize Python script."""
        super().__init__(UnparsePython, tree, source_script, output_dir, "python", "py")

    def write_header(self, output: io.TextIOBase) -> None:
        """Write header of Python script."""
        super().write_header(output)
        output.write("\nimport weechat")

    def write_footer(self, output: io.TextIOBase) -> None:
        """Write footer of Python script."""
        super().write_footer(output)
        output.write('\n\nif __name__ == "__main__":\n    weechat_init()\n')


class WeechatPerlScript(WeechatScript):
    """A WeeChat script written in Perl."""

    def __init__(self, tree: ast.AST, source_script: str, output_dir: str) -> None:
        """Initialize Perl script."""
        super().__init__(UnparsePerl, tree, source_script, output_dir, "perl", "pl")

    def write_footer(self, output: io.TextIOBase) -> None:
        """Write footer of Perl script."""
        super().write_footer(output)
        output.write("\nweechat_init();\n")


class WeechatRubyScript(WeechatScript):
    """A WeeChat script written in Ruby."""

    def __init__(self, tree: ast.AST, source_script: str, output_dir: str) -> None:
        """Initialize Ruby script."""
        super().__init__(UnparseRuby, tree, source_script, output_dir, "ruby", "rb")

    def update_tree(self) -> None:
        """Make changes in AST tree of Ruby script."""
        super().update_tree()
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Attribute) and node.value.id == "weechat":
                node.value.id = "Weechat"
            if (
                isinstance(node, ast.Call)
                and isinstance(node.func, ast.Attribute)
                and node.func.attr == "config_new_option"
            ):
                node.args = [*node.args[:11], ast.List(node.args[11:])]


class WeechatLuaScript(WeechatScript):
    """A WeeChat script written in Lua."""

    def __init__(self, tree: ast.AST, source_script: str, output_dir: str) -> None:
        """Initialize Lua script."""
        super().__init__(UnparseLua, tree, source_script, output_dir, "lua", "lua", comment_char="--")

    def write_footer(self, output: io.TextIOBase) -> None:
        """Write footer of Lua script."""
        super().write_footer(output)
        output.write("\nweechat_init()\n")


class WeechatTclScript(WeechatScript):
    """A WeeChat script written in Tcl."""

    def __init__(self, tree: ast.AST, source_script: str, output_dir: str) -> None:
        """Initialize Tcl script."""
        super().__init__(UnparseTcl, tree, source_script, output_dir, "tcl", "tcl")

    def write_footer(self, output: io.TextIOBase) -> None:
        """Write footer of Tcl script."""
        super().write_footer(output)
        output.write("\nweechat_init\n")


class WeechatGuileScript(WeechatScript):
    """A WeeChat script written in Guile (Scheme)."""

    def __init__(self, tree: ast.AST, source_script: str, output_dir: str) -> None:
        """Initialize Guile script."""
        super().__init__(UnparseGuile, tree, source_script, output_dir, "guile", "scm", comment_char=";")

    def update_tree(self) -> None:
        """Make changes in AST tree of Guile script."""
        super().update_tree()
        functions_with_list = (
            "config_new_section",
            "config_new_option",
        )
        for node in ast.walk(self.tree):
            if (
                isinstance(node, ast.Call)
                and isinstance(node.func, ast.Attribute)
                and node.func.attr in functions_with_list
            ):
                node.args = [ast.Call("list", node.args)]

    def write_footer(self, output: io.TextIOBase) -> None:
        """Write footer of Guile script."""
        super().write_footer(output)
        output.write("\n(weechat_init)\n")


class WeechatJavaScriptScript(WeechatScript):
    """A WeeChat script written in JavaScript."""

    def __init__(self, tree: ast.AST, source_script: str, output_dir: str) -> None:
        """Initialize JavaScript script."""
        super().__init__(UnparseJavaScript, tree, source_script, output_dir, "javascript", "js", comment_char="//")

    def write_footer(self, output: io.TextIOBase) -> None:
        """Writer footer of JavaScript script."""
        super().write_footer(output)
        output.write("\nweechat_init()\n")


class WeechatPhpScript(WeechatScript):
    """A WeeChat script written in PHP."""

    def __init__(self, tree: ast.AST, source_script: str, output_dir: str) -> None:
        """Initialize PHP script."""
        super().__init__(UnparsePhp, tree, source_script, output_dir, "php", "php", comment_char="//")

    def write_header(self, output: io.TextIOBase) -> None:
        """Writer header of PHP script."""
        output.write("<?php\n")
        super().write_header(output)

    def write_footer(self, output: io.TextIOBase) -> None:
        """Write footer of PHP script."""
        super().write_footer(output)
        output.write("\nweechat_init();\n")


# ============================================================================


def update_nodes(tree: ast.AST) -> None:
    """Update the tests AST tree (in-place).

    Actions performed:
      1. add a print message in each test_* function
      2. add arguments in calls to check() function
    """
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) and node.name.startswith("test_"):
            # add a print at the beginning of each test function
            node.body.insert(0, ast.parse(f'weechat.prnt("", "  > {node.name}");'))
        elif isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id == "check":
            # add two arguments in the call to "check" function:
            #   1. the string representation of the test
            #   2. the line number in source (as string)
            # for example if this test is on line 50:
            #   >>> check(weechat.test() == 123)
            # it becomes:
            #   >>> check(weechat.test() == 123, 'weechat.test() == 123', '50')
            output = io.StringIO()
            unparsed = UnparsePython(output=output)
            unparsed.add(node.args[0])
            node.args.append(ast.Constant(output.getvalue()))
            node.args.append(ast.Constant(str(node.func.lineno)))


def get_tests(path: str) -> ast.AST:
    """Parse the source with tests and return the AST node."""
    with Path(path).open() as f:
        test_script = f.read()
        tests = ast.parse(test_script)
        update_nodes(tests)
        return tests


def generate_scripts(source_script: str, output_dir: str) -> Tuple[int, Union[str, None]]:
    """Generate scripts in all languages to test the API."""
    ret_code = 0
    error = None
    try:
        for name, obj in inspect.getmembers(sys.modules[__name__]):
            if (
                inspect.isclass(obj)
                and name != "WeechatScript"
                and name.startswith("Weechat")
                and name.endswith("Script")
            ):
                tests = get_tests(source_script)
                obj(tests, source_script, output_dir).write()
    except Exception as exc:  # noqa: BLE001
        ret_code = 1
        error = f"ERROR: {exc}\n\n{traceback.format_exc()}"
    return ret_code, error


def testapigen_cmd_cb(data: str, buf: str, args: str) -> int:  # noqa: ARG001
    """Callback for WeeChat command /testapigen."""

    def print_error(msg: str) -> None:
        """Print an error message on core buffer."""
        weechat.prnt("", f"{weechat.prefix('error')}{msg}")

    try:
        source_script, output_dir = args.split()
    except ValueError:
        print_error("ERROR: invalid arguments for /testapigen")
        return weechat.WEECHAT_RC_OK
    if not weechat.mkdir_parents(output_dir, 0o755):
        print_error("ERROR: invalid directory: {output_dir}")
        return weechat.WEECHAT_RC_OK
    ret_code, error = generate_scripts(source_script, output_dir)
    if error:
        print_error(error)
    return weechat.WEECHAT_RC_OK if ret_code == 0 else weechat.WEECHAT_RC_ERROR


def get_parser_args() -> argparse.Namespace:
    """Get parser arguments."""
    parser = argparse.ArgumentParser(description=("Generate WeeChat scripts in all languages to test the API."))
    parser.add_argument("script", help="the path to Python script with tests")
    parser.add_argument("-o", "--output-dir", default=".", help="output directory (defaults to current directory)")
    return parser.parse_args()


def main() -> None:
    """Generate all scripts (when script is not loaded in WeeChat)."""
    args = get_parser_args()
    ret_code, error = generate_scripts(args.script, args.output_dir)
    if error:
        print(error)
    sys.exit(ret_code)


if __name__ == "__main__":
    if RUNNING_IN_WEECHAT:
        weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", "")
        weechat.hook_command(
            SCRIPT_COMMAND,
            "Generate scripting API test scripts",
            "source_script output_dir",
            "source_script: path to source script (testapi.py)\n   output_dir: output directory for scripts",
            "",
            "testapigen_cmd_cb",
            "",
        )
    else:
        main()
